@@ -1,5 +1,5 @@
 /* Handle so called `shell archives'.
-   Copyright (C) 1994, 1995, 1996, 2002 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1995, 1996, 2002, 2005 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
 
 #ifdef HAVE_CONFIG_H
@@ -28,6 +28,8 @@
 #include "xalloc.h"
 #include "xgetcwd.h"
 
+#include "scripts.x"
+
 #if HAVE_LOCALE_H
 # include <locale.h>
 #else
@@ -37,6 +39,10 @@
 #define _(str) gettext (str)
 #define N_(str) gettext_noop (str)
 
+#ifndef NUL
+#  define NUL '\0'
+#endif
+
 static const char *cut_mark_line
   = "---- Cut Here and feed the following to sh ----\n";
 
@@ -55,12 +61,6 @@ static const char *cut_mark_line
    character environments get different results.  */
 #define CHARACTER_COUNT_COMMAND "LC_ALL=C wc -c <"
 
-/* Command computing MD5 sumof specified file.  Because there is not
-   official standard for this we have to restrict ourself to a few
-   versions: GNU md5sum from GNU fileutils and the version by
-   Plumb/Lankester.  */
-#define MD5SUM_COMMAND "md5sum"
-
 /* Maximum length for a text line before it is considered binary.  */
 #define MAXIMUM_NON_BINARY_LINE 200
 
@@ -256,10 +256,10 @@ static off_t archive_type_position;
 static off_t first_file_position;
 
 /* Base for output filename.  */
-static char *output_base_name;
+static char *output_base_name = NULL;
 
 /* Actual output filename.  */
-static char *output_filename;
+static char *output_filename = NULL;
 
 static char *submitter_address = NULL;
 
@@ -318,7 +318,6 @@ walkdown (routine, local_name, restore_name)
      const char *restore_name;
 {
   DIR *directory;		/* directory being scanned */
-  struct dirent *entry;		/* current entry in directory */
   int status;			/* status to return */
 
   char *local_name_copy;	/* writeable copy of local_name */
@@ -326,6 +325,7 @@ walkdown (routine, local_name, restore_name)
   size_t sizeof_local_name;	/* allocated size of local_name_copy */
 
   char *restore_name_copy;	/* writeable copy of restore_name */
+  int    restore_offset;	/* passdown copy of restore_name */
   size_t restore_name_length;	/* number of characters in restore_name_copy */
   size_t sizeof_restore_name;	/* allocated size of restore_name_copy */
 
@@ -346,57 +346,76 @@ walkdown (routine, local_name, restore_name)
 
   status = 0;
 
-  local_name_copy = xstrdup (local_name);
-  local_name_length = strlen (local_name_copy);
-  sizeof_local_name = local_name_length + 1;
+  /* include trailing '/' in length */
 
-  restore_name_copy = xstrdup (restore_name);
-  restore_name_length = strlen (restore_name_copy);
-  sizeof_restore_name = restore_name_length + 1;
+  local_name_length = strlen (local_name) + 1;
+  sizeof_local_name = local_name_length + 32;
+  local_name_copy   = xmalloc (sizeof_local_name);
+  memcpy (local_name_copy, local_name, local_name_length-1);
+  local_name_copy[ local_name_length-1 ] = '/';
+  local_name_copy[ local_name_length   ] = NUL;
 
-  while (!status && (entry = readdir (directory), entry))
-    if (strcmp (entry->d_name, ".") && strcmp (entry->d_name, ".."))
-      {
-	int added_size = 1 + NAMLEN (entry);
+  restore_name_length = strlen (restore_name) + 1;
+  sizeof_restore_name = restore_name_length + 32;
+  restore_name_copy   = xmalloc (sizeof_restore_name);
+  memcpy (restore_name_copy, restore_name, restore_name_length-1);
+  restore_name_copy[ restore_name_length-1 ] = '/';
+  restore_name_copy[ restore_name_length   ] = NUL;
 
-	/* Update file names, reallocating them as required.  */
+  if ((restore_name_copy[0] == '.') && (restore_name_copy[1] == '/'))
+    restore_offset = 2;
+  else
+    restore_offset = 0;
 
-	if (local_name_length + added_size + 1 > sizeof_local_name)
-	  {
-	    sizeof_local_name = local_name_length + added_size + 1;
-	    local_name_copy = (char *)
-	      xrealloc (local_name_copy, sizeof_local_name);
-	  }
-	sprintf (local_name_copy + local_name_length, "/%s", entry->d_name);
+  for (;;)
+    {
+      struct dirent *entry = readdir (directory);
+      const char* pzN;
+      int space_need;
 
-	if (restore_name_length + added_size + 1 > sizeof_restore_name)
-	  {
-	    sizeof_restore_name = restore_name_length + added_size + 1;
-	    restore_name_copy = (char *)
-	      xrealloc (restore_name_copy, sizeof_restore_name);
-	  }
-	sprintf (restore_name_copy + restore_name_length, "/%s",
-		 entry->d_name);
+      if (entry == NULL)
+	break;
 
-	/* Avoid restoring `./xxx' when shar'ing `.'.  */
+      /* append the new file name after the trailing '/' char.
+         If we need more space, add in a buffer so we needn't
+         allocate over and over.  */
 
-	if (!strncmp (restore_name_copy, "./", 2))
-	  {
-	    const char *rp = restore_name_copy + 2;
-	    char *wp = restore_name_copy;
-	    while ((*wp++ = *rp++));
-	    restore_name_length -= 2;
-	  }
+      pzN = entry->d_name;
+      if (*pzN == '.')
+	{
+	  if (pzN[1] == NUL)
+	    continue;
+	  if ((pzN[1] == '.') && (pzN[2] == NUL))
+	    continue;
+	}
 
-	status = walkdown (routine, local_name_copy, restore_name_copy);
-      }
+      space_need = 1 + NAMLEN (entry);
+      if (local_name_length + space_need > sizeof_local_name)
+	{
+	  sizeof_local_name = local_name_length + space_need + 16;
+	  local_name_copy = (char *)
+	    xrealloc (local_name_copy, sizeof_local_name);
+	}
+      strcpy (local_name_copy + local_name_length, pzN);
+
+      if (restore_name_length + space_need > sizeof_restore_name)
+	{
+	  sizeof_restore_name = restore_name_length + space_need + 16;
+	  restore_name_copy = (char *)
+	    xrealloc (restore_name_copy, sizeof_restore_name);
+	}
+      strcpy (restore_name_copy + restore_name_length, pzN);
+
+      status = walkdown (routine, local_name_copy,
+			 restore_name_copy + restore_offset);
+      if (status != 0)
+	break;
+    }
 
   /* Clean up.  */
 
-  if (sizeof_local_name > 0)
-    free (local_name_copy);
-  if (sizeof_restore_name > 0)
-    free (restore_name_copy);
+  free (local_name_copy);
+  free (restore_name_copy);
 
 #if CLOSEDIR_VOID
   closedir (directory);
@@ -425,18 +444,19 @@ walktree (routine, local_name)
 {
   const char *restore_name;
   char *local_name_copy;
-  char *cursor;
-  int status;
-  int len;
 
   /* Remove crumb at end.  */
+  {
+    int len = strlen (local_name);
+    char *cursor;
 
-  len = strlen (local_name);
-  local_name_copy = (char *) alloca (len + 1);
-  memcpy (local_name_copy, local_name, len + 1);
-  cursor = local_name_copy + len - 1;
-  while (*cursor == '/' && cursor > local_name_copy)
-    *cursor-- = '\0';
+    local_name_copy = (char *) alloca (len + 1);
+    memcpy (local_name_copy, local_name, len + 1);
+    cursor = local_name_copy + len - 1;
+
+    while (*cursor == '/' && cursor > local_name_copy)
+      *(cursor--) = NUL;
+  }
 
   /* Remove crumb at beginning.  */
 
@@ -451,21 +471,24 @@ walktree (routine, local_name)
 
   /* Just act on current entry.  */
 
-  status = stat (local_name_copy, &struct_stat);
-  if (status != 0)
-    error (0, errno, local_name_copy);
-  else
-    status = (*routine) (local_name_copy, restore_name);
+  {
+    int status = stat (local_name_copy, &struct_stat);
+
+    if (status != 0)
+      error (0, errno, local_name_copy);
+    else
+      status = (*routine) (local_name_copy, restore_name);
+
+    return status;
+  }
 
 #else
 
   /* Walk recursively.  */
 
-  status = walkdown (routine, local_name_copy, restore_name);
+  return walkdown (routine, local_name_copy, restore_name);
 
 #endif
-
-  return status;
 }
 
 /* Generating parts of shar file.  */
@@ -516,43 +539,15 @@ mode_string (mode)
 static void
 generate_configure ()
 {
+  if (md5_count_mode)
+    fprintf (output, md5check_z, N_("\
+Note: not verifying md5sums.  Consider installing GNU coreutils."));
+
   if (no_i18n)
     fputs ("echo=echo\n", output);
   else
     {
-      fprintf (output, "\
-save_IFS=\"${IFS}\"\n\
-IFS=\"${IFS}:\"\n\
-gettext_dir=FAILED\n\
-locale_dir=FAILED\n\
-first_param=\"$1\"\n\
-for dir in $PATH\n\
-do\n\
-  if test \"$gettext_dir\" = FAILED && test -f $dir/gettext \\\n\
-     && ($dir/gettext --version >/dev/null 2>&1)\n\
-  then\n\
-    case `$dir/gettext --version 2>&1 | sed 1q` in\n\
-      *GNU*) gettext_dir=$dir ;;\n\
-    esac\n\
-  fi\n\
-  if test \"$locale_dir\" = FAILED && test -f $dir/%s \\\n\
-     && ($dir/%s --print-text-domain-dir >/dev/null 2>&1)\n\
-  then\n\
-    locale_dir=`$dir/%s --print-text-domain-dir`\n\
-  fi\n\
-done\n\
-IFS=\"$save_IFS\"\n\
-if test \"$locale_dir\" = FAILED || test \"$gettext_dir\" = FAILED\n\
-then\n\
-  echo=echo\n\
-else\n\
-  TEXTDOMAINDIR=$locale_dir\n\
-  export TEXTDOMAINDIR\n\
-  TEXTDOMAIN=sharutils\n\
-  export TEXTDOMAIN\n\
-  echo=\"$gettext_dir/gettext -s\"\n\
-fi\n",
-	       "shar", "shar", "shar");
+      fputs (i18n_z, output);
       /* Above the name of the program of the package which supports the
 	 --print-text-domain-dir option has to be given.  */
     }
@@ -599,40 +594,10 @@ fi\n",
 
   if (timestamp_mode)
     {
-      const char *file = "$$.touch";
-      const char *stamp1 = "200112312359.59";
-      const char *stamp2 = "123123592001.59";
-      const char *stamp2tr = "123123592001.5"; /* old SysV 14-char limit */
-      const char *stamp3 = "1231235901";
-
-      /* Set the shell variable shar_touch to an appropriate invocation
-	 of `touch' if the touch program is proven able to restore dates.
-	 Otherwise, set shar_touch to `:' and issue a warning.  */
-
-      fprintf (output, "\
-if touch -am -t %s %s >/dev/null 2>&1 && test ! -f %s && test -f %s; then\n\
-  shar_touch='touch -am -t $1$2$3$4$5$6.$7 \"$8\"'\n\
-elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s && test ! -f %s && test -f %s; then\n\
-  shar_touch='touch -am $3$4$5$6$1$2.$7 \"$8\"'\n\
-elif touch -am %s %s >/dev/null 2>&1 && test ! -f %s && test -f %s; then\n\
-  shar_touch='touch -am $3$4$5$6$2 \"$8\"'\n\
-else\n\
-  shar_touch=:\n\
-  echo\n\
-  $echo '%s'\n\
-  $echo \"%s\"\n\
-  echo\n\
-fi\n\
-rm -f %s %s %s %s %s\n\
-#\n",
-	       stamp1, file, stamp1, file,
-	       stamp2, file, stamp2, stamp2tr, file,
-	       stamp3, file, stamp3, file,
-	       N_("\
+      fprintf (output, timestamp_z, N_("\
 WARNING: not restoring timestamps.  Consider getting and"),
 	       N_("\
-installing GNU \\`touch', distributed in GNU File Utilities..."),
-	       stamp1, stamp2, stamp2tr, stamp3, file);
+installing GNU `touch'\\'', distributed in GNU File Utilities..."));
     }
 
   if (file_size_limit == 0 || part_number == 1)
@@ -656,34 +621,56 @@ fi\n",
     }
 }
 
-/*---.
-| ?  |
-`---*/
+/*----------------------------------------------.
+| generate_mkdir                                |
+| Make sure it is done only once for each dir   |
+`----------------------------------------------*/
 
-/* Ridiculously enough.  FIXME: No fixed limit in GNU...  */
-#define MAX_MKDIR_ALREADY	128
-
-char *mkdir_already[MAX_MKDIR_ALREADY];
-int mkdir_already_count = 0;
+static int    mkdir_alloc_ct = 0;
+static int    mkdir_already_ct = 0;
+static char** mkdir_already;
 
 static void
 generate_mkdir (path)
      const char *path;
 {
-  int counter;
-  char *cursor;
-
   /* If already generated code for this dir creation, don't do again.  */
 
-  for (counter = 0; counter < mkdir_already_count; counter++)
-    if (!strcmp (path, mkdir_already[counter]))
-      return;
+  {
+    int    ct = mkdir_already_ct;
+    char** pp = mkdir_already;
+
+    while (--ct > 0)
+      {
+        if (strcmp (*(pp++), path) == 0)
+          return;
+      }
+  }
 
   /* Haven't done this one.  */
 
-  if (mkdir_already_count == MAX_MKDIR_ALREADY)
-    error (EXIT_FAILURE, 0, _("Too many directories for mkdir generation"));
-  cursor = mkdir_already[mkdir_already_count++] = xstrdup (path);
+  if (++mkdir_already_ct > mkdir_alloc_ct)
+    {
+      /*
+       *  We need more name space.  Get larger and larger chunks of space.
+       *  The bound is when integers go negative.  Too many directories.  :)
+       *
+       *  16, 40, 76, 130, 211, 332, 514, 787, 1196, 1810, 2731, 4112, ...
+       */
+      mkdir_alloc_ct += 16 + (mkdir_alloc_ct/2);
+      if (mkdir_alloc_ct < 0)
+        error (EXIT_FAILURE, 0, _("Too many directories for mkdir generation"));
+
+      if (mkdir_already != NULL)
+        mkdir_already =
+          xrealloc (mkdir_already, mkdir_alloc_ct * sizeof (char*));
+      else
+        mkdir_already = xmalloc (mkdir_alloc_ct * sizeof (char*));
+    }
+
+  /* Add the directory into our "we've done this already" table */
+
+  mkdir_already[ mkdir_already_ct-1 ] = xstrdup (path);
 
   /* Generate the text.  */
 
@@ -692,12 +679,26 @@ generate_mkdir (path)
     fprintf (output, "\
   $echo $echo_n 'x -' '%s: '$echo_c\n\
   if mkdir '%s'; then $echo '%s'; else $echo '%s'; fi\n",
-	     path, path, N_("created"), N_("failed to create"));
+             path, path, N_("created"), N_("failed to create"));
   else
     fprintf (output, "  mkdir '%s'\n", path);
   fputs ("fi\n", output);
 }
 
+static void
+clear_mkdir_already (void)
+{
+  char** pp = mkdir_already;
+  int    ct = mkdir_already_ct;
+
+  mkdir_already_ct = 0;
+  while (--ct >= 0)
+    {
+      free (*pp);
+      *(pp++) = NULL;
+    }
+}
+
 /*---.
 | ?  |
 `---*/
@@ -1015,8 +1016,7 @@ exit 0\n",
 
 	  /* Clear mkdir_already in case the user unshars out of order.  */
 
-	  while (mkdir_already_count > 0)
-	    free (mkdir_already[--mkdir_already_count]);
+	  clear_mkdir_already ();
 
 	  /* Form the next filename.  */
 
@@ -1626,21 +1626,15 @@ else\n",
 	  size_t cnt;
 	  did_md5 = 1;
 
-	  fprintf (output, "\
-  if ( %s --help </dev/null 2>&1 | grep 'sage: md5sum \\[' ) >/dev/null 2>&1 \\\n\
-  && ( %s --version </dev/null 2>&1 | grep -v 'textutils 1.12' ) >/dev/null; then\n\
-    %s -c << %s >/dev/null 2>&1 \\\n\
-    || $echo '%s:' '%s'\n",
-		   MD5SUM_COMMAND, MD5SUM_COMMAND, MD5SUM_COMMAND,
-		   DEFAULT_HERE_DELIMITER, restore_name,
-		   N_("MD5 check failed"));
+	  fprintf (output, md5test_z, restore_name,
+		   N_("MD5 check failed"), here_delimiter);
 
 	  for (cnt = 0; cnt < 16; ++cnt)
 	    fprintf (output, "%02x", md5buffer[cnt]);
 
 	  fprintf (output, " %c%s\n\
 %s\n",
-		   ' ', restore_name, DEFAULT_HERE_DELIMITER);
+		   ' ', restore_name, here_delimiter);
 	  /* This  ^^^ space is not necessarily a parameter now.  But it
 	     is a flag for binary/text mode and will perhaps be used later.  */
 	}
@@ -1678,14 +1672,14 @@ else\n",
 			break;
 		      }
                   }
-                if (pz[-1] != '\0') do
+                if (pz[-1] != NUL) do
                   {
                     int ch = (unsigned int)fgetc (pfp);
                     if ((ch < 0) || isspace (ch))
                       break;
                     *pz = ch;
                   } while (++pz < wc+BUFSIZ-1);
-                *pz = '\0';
+                *pz = NUL;
               }
 
 	      fprintf (output, "\
@@ -1747,13 +1741,12 @@ set_file_mode (mode)
 static void
 open_output ()
 {
-  size_t l;
-  l = strlen(output_base_name) + 128;
-  if (output_filename)
-    free(output_filename);
-  output_filename = xmalloc(l);
-  snprintf(output_filename, l, output_base_name, ++part_number);
+  if (output_filename == NULL)
+    error (EXIT_FAILURE, ENXIO, _("allocating output file name"));
+
+  sprintf(output_filename, output_base_name, ++part_number);
   output = fopen (output_filename, "w");
+
   if (!output)
     error (EXIT_FAILURE, errno, _("Opening `%s'"), output_filename);
 }
@@ -1926,44 +1919,65 @@ set_file_size_limit (char const *limit)
   char *numend;
   intmax_t lim = strtoimax (limit, &numend, 10);
   if (*numend || ! (0 < lim && lim <= TYPE_MAXIMUM (off_t) / 1024))
-    error (EXIT_FAILURE, 0, "invalid file size limit `%s'", limit);
+    error (EXIT_FAILURE, 0, _("invalid file size limit `%s'"), limit);
   lim *= 1024;
   if (! file_size_limit || lim < file_size_limit)
     file_size_limit = lim;
 }
 
 
-char *parse_output_base_name(char *arg)
+char *
+parse_output_base_name(char *arg)
 {
   int c;
   int hadarg = 0;
   char *fmt, *p;
+  int base_name_len = 128;
 
   for (p = arg ; (c = *p++) != 0; )
     {
+      base_name_len++;
       if (c != '%')
 	continue;
       c = *p++;
       if (c == '%')
 	continue;
       if (hadarg)
-	return 0;
+	return NULL;
       while (c != 0 && strchr("#0+- 'I", c) != 0)
 	c = *p++;
-      while (c != 0 && c >= '0' && c <= '9')
-	c = *p++;
+      if (c == 0)
+	return NULL;
+      if (c >= '0' && c <= '9')
+	{
+	  long v;
+	  errno = 0;
+	  v = strtol(p-1, &fmt, 10);
+	  if ((v == 0) || (v > 16) || (errno != 0))
+	    {
+	      fprintf(stderr, _("invalid format (count field too wide): '%s'\n"),
+		      arg);
+	      return NULL;
+	    }
+	  p = fmt;
+	  c = *p++;
+	  base_name_len += v;
+	}
       if (c == '.')
-	c = *p++;
-      while (c != 0 && c >= '0' && c <= '9')
-	c = *p++;
+	{
+	  c = *p++;
+	  while (c != 0 && c >= '0' && c <= '9')
+	    c = *p++;
+	}
       if (c == 0 || strchr("diouxX", c) == 0)
-	return 0;
+	return NULL;
       hadarg = 1;
     }
   fmt = xmalloc(strlen(arg) + (hadarg ? 1 : 6));
   strcpy(fmt, arg);
   if (!hadarg)
     strcat(fmt, ".%02d");
+  output_filename = xmalloc(base_name_len);
   return fmt;
 }
 
@@ -1995,7 +2009,7 @@ main (argc, argv)
 	 optchar != EOF)
     switch (optchar)
       {
-      case '\0':
+      case NUL:
 	break;
 
       case '$':
@@ -2108,8 +2122,6 @@ This system doesn't support -Z ('compress'), use -z instead"));
 	break;
 
       case 'o':
-	if (output_base_name)
-	  free (output_base_name);
         output_base_name = parse_output_base_name(optarg);
         if (!output_base_name)
 	  {
@@ -2261,7 +2273,7 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 	  if (argc == max_argc)
 	    list = (char **) xrealloc (list,
 				       (max_argc *= 2) * sizeof (char *));
-	  if (stdin_buf[0] != '\0')
+	  if (stdin_buf[0] != NUL)
 	    stdin_buf[strlen (stdin_buf) - 1] = 0;
 	  list[argc] = xstrdup (stdin_buf);
 	  ++argc;
@@ -2302,8 +2314,7 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 	error (0, 0, _("PLEASE avoid -X shars on Usenet or public networks"));
 
       fputs ("\
-shar_wish=\n",
-	     output);
+shar_wish=\n", output);
     }
 
   first_file_position = ftello (output);
@@ -2311,28 +2322,29 @@ shar_wish=\n",
   /* Process positional parameters and files.  */
 
   for (; optind < argc; optind++)
-    if (intermixed_parameter_mode)
-      if (strcmp (argv[optind], "-B") == 0)
-	set_file_mode ('B');
-      else if (strcmp (argv[optind], "-T") == 0)
-	set_file_mode ('T');
-      else if (strcmp (argv[optind], "-M") == 0)
-	set_file_mode ('M');
-      else if (strcmp (argv[optind], "-z") == 0)
-	set_file_mode ('z');
-      else if (strcmp (argv[optind], "-Z") == 0
-	       || strcmp (argv[optind], "-C") == 0)
-	set_file_mode ('Z');
-      else
-	{
-	  if (walktree (shar, argv[optind]))
-	    status = EXIT_FAILURE;
-	}
-    else
-      {
-	if (walktree (shar, argv[optind]))
-	  status = EXIT_FAILURE;
-      }
+    {
+      char* arg = argv[optind];
+
+      if (intermixed_parameter_mode && (arg[0] == '-') && (arg[2] == NUL))
+        {
+          switch (arg[1]) {
+          case 'B':
+          case 'T':
+          case 'M':
+          case 'z':
+          case 'Z':
+            set_file_mode (arg[1]);
+            continue;
+
+          case 'C':
+            set_file_mode ('Z');
+            continue;
+          }
+        }
+
+      if (walktree (shar, arg))
+        status = EXIT_FAILURE;
+    }
 
   /* Delete the sequence file, if any.  */
 
